"""
"""
import numpy as np

from .reduction import Reduction


class POD(Reduction):
    def __init__(self, method='svd', **kwargs):
        """
        """
        available_methods = {
            'svd': (self._svd, {
                'rank': -1
            }),
            'randomized_svd': (self._rsvd, {
                'rank': -1
            }),
            'correlation_matrix': (self._corrm, {
                'rank': -1,
                'save_memory': False
            }),
        }

        self._modes = None
        self._singular_values = None

        method = available_methods.get(method)
        if method is None:
            raise RuntimeError(
                "Invalid method for POD. Please chose one among {}".format(
                    ', '.join(available_methods)))

        self.__method, default_args = method
        kwargs.update(default_args)

        for hyperparam, value in kwargs.items():
            setattr(self, hyperparam, value)

    @property
    def modes(self):
        """
        The POD modes.

        :type: numpy.ndarray
        """
        return self._modes

    @property
    def singular_values(self):
        """
        The singular values  

        :type: numpy.ndarray
        """
        return self._singular_values

    def reduce(self, X):
        """
        Reduces the parameter Space by using the specified reduction method (default svd).

        :type: numpy.ndarray
        """
        self._modes, self._singular_values = self.__method(X)
        return self.modes.T.dot(X)

    def expand(self, X):
        """
        Projects a reduced to full order solution.

        :type: numpy.ndarray
        """
        return np.sum(self.modes * X, axis=1)

    def _truncation(self, X):
        """
        Return the number of modes to select according to the `rank` parameter.
        See POD.__init__ for further info.

        :return: the number of modes
        :rtype: int
        """
        if self.rank is 0:
            omega = lambda x: 0.56 * x**3 - 0.95 * x**2 + 1.82 * x + 1.43
            beta = np.divide(*sorted(X.shape))
            tau = np.median(s) * omega(beta)
            rank = np.sum(s > tau)
        elif self.rank > 0 and svd_rank < 1:
            cumulative_energy = np.cumsum(s**2 / (s**2).sum())
            rank = np.searchsorted(cumulative_energy, svd_rank) + 1
        elif self.rank >= 1 and isinstance(svd_rank, int):
            rank = min(svd_rank, U.shape[1])
        else:
            rank = X.shape[1]

        return rank

    def _svd(self, X):
        """
        Truncated Singular Value Decomposition.

        :param numpy.ndarray X: the matrix to decompose.
        :return: the truncated left-singular vectors matrix, the truncated
            singular values array, the truncated right-singular vectors matrix.
        :rtype: numpy.ndarray, numpy.ndarray, numpy.ndarray

        References:
        Gavish, Matan, and David L. Donoho, The optimal hard threshold for
        singular values is, IEEE Transactions on Information Theory 60.8
        (2014): 5040-5053.
        """
        U, s = np.linalg.svd(X, full_matrices=False)[:2]

        rank = self._truncation(X)
        return U[:, :rank], s[:rank]

    def _rsvd(self, X):
        """
        Truncated randomized Singular Value Decomposition.
        
        :param numpy.ndarray X: the matrix to decompose.
        :return: the truncated left-singular vectors matrix, the truncated
            singular values array, the truncated right-singular vectors matrix.
        :rtype: numpy.ndarray, numpy.ndarray, numpy.ndarray
        """

        P = np.random.rand(X.shape[1], X.shape[0])
        Q = np.linalg.qr(X.dot(P))[0]

        Y = Q.T.conj().dot(X)

        Uy, s = np.linalg.svd(Y, full_matrices=False)[:2]
        U = Q.dot(Uy)

        rank = self._truncation(X)
        return U[:, :rank], s[:rank]

    def _corrm(self, X):
        """
        Truncated Singular Value Decomposition. calculated with correlation matrix.
        
        :param numpy.ndarray X: the matrix to decompose.
        :return: the truncated left-singular vectors matrix, the truncated
            singular values array, the truncated right-singular vectors matrix.
        :rtype: numpy.ndarray, numpy.ndarray, numpy.ndarray
        """

        if self.save_memory:
            corr = np.empty(size=(X.shape[1], X.shape[1]))
            for i, i_snap in enumerate(X.T):
                for j, k_snap in enumerate(X.T):
                    corr[i, j] = np.inner(i_snap, k_snap)

        else:
            corr = X.T.dot(X)

        s, U = np.linalg.eig(corr)
        U = X.dot(U) / np.sqrt(s)
        rank = self._truncation(X)

        return U[:, :rank], s[:rank]



__all__ = [
    'database',
    'reduction', 'pod',
    'approximation', 'rbf', 'linear',

]

__project__ = 'EZyRB'
__title__ = "ezyrb"
__author__ = "Nicola Demo, Marco Tezzele"
__copyright__ = "Copyright 2016-2019, EZyRB contributors"
__license__ = "MIT"
__version__ = "1.0"
__mail__ = 'demo.nicola@gmail.com, marcotez@gmail.com'
__maintainer__ = __author__
__status__ = "Stable"

from .database import Database
from .reduction import Reduction
from .pod import POD
from .rbf import RBF
from .reducedordermodel import ReducedOrderModel



"""
Module for Linear N-Dimensional Interpolation
"""

import numpy as np
from scipy.interpolate import LinearNDInterpolator

from .approximation import Approximation


class Linear(Approximation):
    """
    Multidimensional linear interpolator.

    :param float fill_value: value used to fill in for requested points outside
        of the convex hull of the input points. If not provided, then the
        default is numpy.nan.
    """

    def __init__(self, fill_value=np.nan):
        self.fill_value = fill_value

    def fit(self, points, values):
        """
        Construct the interpolator given `points` and `values`.

        :param array_like points: the coordinates of the points.
        :param array_like values: the values in the points.
        """
        self.intepolator = LinearNDInterpolator(points, values)

    def predict(self, new_point):
        """
        Evaluate interpolator at given `new_points`.

        :param array_like new_points: the coordinates of the given points.
        :return: the interpolated values.
        :rtype: numpy.ndarray
        """
        return self.interpolator(new_point)



"""
Module for the snapshots database collected during the Offline stage
"""
import numpy as np


class Database(object):
    """
    Database class

    :param array_like parameters: the input parameters 
    :param array_like snapshots: the input snapshots
    :param Scale scaler_parameters: the scaler for the parameters. Default
        is None meaning no scaling.
    :param Scale scaler_snapshots: the scaler for the snapshots. Default is
        None meaning no scaling.
    """
    def __init__(self,
                 parameters=None,
                 snapshots=None,
                 scaler_parameters=None,
                 scaler_snapshots=None):
        self._parameters = None
        self._snapshots = None
        self.scaler_parameters = scaler_parameters
        self.scaler_snapshots = scaler_snapshots

        # if only parameters or snapshots are provided
        if (parameters is None) ^ (snapshots is None):
            raise RuntimeError

        if not (parameters is None) and not (snapshots is None):
            self.add(parameters, snapshots)

    @property
    def parameters(self):
        """
        The matrix containing the input parameters (by row).

        :rtype: numpy.ndarray
        """
        if self.scaler_parameters:
            return self.scaler_parameters.fit_transform(self._parameters)
        else:
            return self._parameters

    @property
    def snapshots(self):
        """
        The matrix containing the snapshots (by row).

        :rtype: numpy.ndarray
        """
        if self.scaler_snapshots:
            return self.scaler_snapshots.fit_transform(self._snapshots)
        else:
            return self._snapshots

    def __getitem__(self, val):
        """
        This method returns a new Database with the selected parameters and snapshots.

        .. warning:: The new parameters and snapshots are a view of the
            original Database.
        """
        return Database(self._parameters[val], self._snapshots[val],
                        self.scaler_parameters, self.scaler_snapshots)

    def __len__(self):
        return len(self._snapshots)

    def add(self, parameters, snapshots):
        """
        """

        if len(parameters) is not len(snapshots):
            raise RuntimeError('Different number of parameters and snapshots.')

        if self._parameters is None and self._snapshots is None:
            self._parameters = parameters
            self._snapshots = snapshots
        else:
            self._parameters = np.vstack([self._parameters, parameters])
            self._snapshots = np.vstack([self._snapshots, snapshots])

        return self



"""
Module for the Reduction abstract class
"""

from abc import ABC, abstractmethod


class Reduction(ABC):
    @abstractmethod
    def reduce(self):
        pass

    @abstractmethod
    def expand(self):
        pass



"""
Module for the Approximation abstract class
"""
from abc import ABC, abstractmethod

class Approximation(ABC):
    @abstractmethod
    def fit(self, points, values):
        pass

    @abstractmethod
    def predict(self, new_point):
        pass



"""
Module for Radial Basis Function Interpolation
"""

import numpy as np
from scipy.interpolate import Rbf

from .approximation import Approximation


class RBF(Approximation):
    """
    Multidimensional interpolator using Radial Basis Function.

    :param kernel: The radial basis function; the default is ‘multiquadric’.
    :type kernel: str or callable
    :param float smooth: values greater than zero increase the smoothness of
        the approximation. 0 is for interpolation (default), the function will
        always go through the nodal points in this case.

    :cvar kernel: The radial basis function; the default is ‘multiquadric’.
    :cvar list interpolators: the RBF interpolators (the number of
        interpolators depenend by the dimensionality of the output)
    """

    def __init__(self, kernel='multiquadric', smooth=0):
        self.kernel = kernel
        self.smooth = smooth

    def fit(self, points, values):
        """
        Construct the interpolator given `points` and `values`.

        :param array_like points: the coordinates of the points.
        :param array_like values: the values in the points.
        """
        self.interpolators = []
        for value in values:
            argument = np.hstack([points, value.reshape(-1, 1)]).T
            self.interpolators.append(
                Rbf(*argument, smooth=self.smooth, function=self.kernel))

    def predict(self, new_point):
        """
        Evaluate interpolator at given `new_points`.

        :param array_like new_points: the coordinates of the given points.
        :return: the interpolated values.
        :rtype: numpy.ndarray
        """
        return np.array([interp(*new_point) for interp in self.interpolators])



from setuptools import setup, Command
import os
import sys
import ezyrb
from shutil import rmtree

# Package meta-data.
NAME = ezyrb.__title__
DESCRIPTION = 'Easy Reduced Basis'
URL = 'https://github.com/mathLab/EZyRB'
MAIL = ezyrb.__mail__
AUTHOR = ezyrb.__author__
VERSION = ezyrb.__version__
KEYWORDS='pod interpolation reduced-basis model-order-reduction'

REQUIRED = [
    'future', 'numpy', 'scipy',	'matplotlib',
]

EXTRAS = {
    'docs': ['Sphinx==1.4', 'sphinx_rtd_theme'],
}

LDESCRIPTION = (
    "EZyRB is a python library for the Model Order Reduction based on "
    "baricentric triangulation for the selection of the parameter points and on "
    "Proper Orthogonal Decomposition for the selection of the modes. It is "
    "ideally suited for actual industrial problems, since its structure can "
    "interact with several simulation software simply providing the output file "
    "of the simulations."
)

here = os.path.abspath(os.path.dirname(__file__))
class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds...')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution...')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine...')
        os.system('twine upload dist/*')

        self.status('Pushing git tags...')
        os.system('git tag v{0}'.format(VERSION))
        os.system('git push --tags')

        sys.exit()

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LDESCRIPTION,
    author=AUTHOR,
    author_email=MAIL,
	classifiers=[
	  	'Development Status :: 5 - Production/Stable',
	  	'License :: OSI Approved :: MIT License',
	  	'Programming Language :: Python :: 3.5',
	  	'Programming Language :: Python :: 3.6',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics'
	],
	keywords=KEYWORDS,
	url=URL,
	license='MIT',
	packages=[NAME],
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    test_suite='nose.collector',
	tests_require=['nose'],
	include_package_data=True,
	zip_safe=False,

    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },)



"""
Module for the Reduced Order Modeling
"""
import numpy as np
import math
from scipy.spatial import Delaunay


class ReducedOrderModel(object):
    def __init__(self, database, reduction, approximation):
        self.database = database
        self.reduction = reduction
        self.approximation = approximation

    def fit(self):
        """
        Calculate reduced space
        """
        self.approximation.fit(self.database.parameters,
                               self.reduction.reduce(self.database.snapshots.T))

        return self

    def predict(self, mu):
        """
        Calculate predicted solution for given mu
        """
        return self.reduction.expand(self.approximation.predict(mu))

    def loo_error(self, norm=np.linalg.norm):
        """
        Estimate the approximation error using *leave-one-out* strategy. The
        main idea is to create several reduced spaces by combining all the
        snapshots except one. The error vector is computed as the difference
        between the removed snapshot and the projection onto the properly
        reduced space. The procedure repeats for each snapshot in the database.
        The `func` is applied on each vector of error to obtained a float
        number.

        :param function func: the function used to assign at each vector of
            error a float number. It has to take as input a 'numpy.ndarray` and
            returns a float. Default value is the L2 norm.
        :return: the vector that contains the errors estimated for all
            parametric points.
        :rtype: numpy.ndarray
        """
        error = np.zeros(len(self.database))
        db_range = list(range(len(self.database)))

        for j in db_range:

            remaining_index = db_range[:]
            remaining_index.remove(j)
            new_db = self.database[remaining_index]
            rom = ReducedOrderModel(new_db, self.reduction,
                    self.approximation).fit()

            error[j] = norm(self.database.snapshots[j] -
                            rom.predict(self.database.parameters[j]))

        return error

    def optimal_mu(self, error=None, k=1):
        """
        Return the parametric points where new high-fidelity solutions have to
        be computed in ordere to globaly reduce the estimated error. These
        points are the barycentric center of the region (simplex) with higher
        error.

        :param numpy.ndarray error: the estimated error evaluated for each
            snapshot; if error array is not passed, it is computed using
            :func:`loo_error` with the default function. Default value is None.
        :param int k: the number of optimal points to return. Default value is
            1.
        :return: the optimal points
        :rtype: list(numpy.ndarray)
        """
        if error is None:
            error = self.loo_error()

        mu = self.database.parameters
        tria = Delaunay(mu)

        error_on_simplex = np.array([
            np.sum(error[smpx]) * self._simplex_volume(mu[smpx])
            for smpx in tria.simplices
        ])

        barycentric_point = []
        for index in np.argpartition(error_on_simplex, -k)[-k:]:
            worst_tria_pts = mu[tria.simplices[index]]
            worst_tria_err = error[tria.simplices[index]]

            barycentric_point.append(
                np.average(worst_tria_pts, axis=0, weights=worst_tria_err))

        return barycentric_point

    def _simplex_volume(self, vertices):
        """
         Method implementing the computation of the volume of a N dimensional
         simplex.
         Source from: `wikipedia.org/wiki/Simplex
         <https://en.wikipedia.org/wiki/Simplex>`_.
         :param numpy.ndarray simplex_vertices: Nx3 array containing the
             parameter values representing the vertices of a simplex. N is the
             dimensionality of the parameters.
         :return: N dimensional volume of the simplex.
         :rtype: float
         """
        distance = np.transpose([vertices[0] - vi for vi in vertices[1:]])
        return np.abs(
            np.linalg.det(distance) / math.factorial(vertices.shape[1]))
